'use strict';

// //////////////////////////////////////////////////////////////////////////////
// DISCLAIMER
//
// Copyright 2016 ArangoDB GmbH, Cologne, Germany
//
// The Programs (which include both the software and documentation) contain
// proprietary information of ArangoDB GmbH; they are provided under a license
// agreement containing restrictions on use and disclosure and are also
// protected by copyright, patent and other intellectual and industrial
// property laws. Reverse engineering, disassembly or decompilation of the
// Programs, except to the extent required to obtain interoperability with
// other independently created software or as specified by law, is prohibited.
//
// It shall be the licensee's responsibility to take all appropriate fail-safe,
// backup, redundancy, and other measures to ensure the safe use of
// applications if the Programs are used for purposes such as nuclear,
// aviation, mass transit, medical, or other inherently dangerous applications,
// and ArangoDB GmbH disclaims liability for any damages caused by such use of
// the Programs.
//
// This software is the confidential and proprietary information of ArangoDB
// GmbH. You shall not disclose such confidential and proprietary information
// and shall use it only in accordance with the terms of the license agreement
// you entered into with ArangoDB GmbH.
//
// @author Michael Hackstein
// //////////////////////////////////////////////////////////////////////////////

const arangodb = require('@arangodb');
const ArangoError = arangodb.ArangoError;
const errors = arangodb.errors;
const db = require('internal').db;
const gm = require('@arangodb/general-graph-common');

const usageMessage = 'A SmartGraph info needs to have the following format: {numberOfShards: <integer>, smartGraphAttribute: <string>, ?edgeDefinitions: [], ?orphanCollections: [<string>]}';

const getGraphCollection = () => {
  const gCol = db._collection('_graphs');
  if (gCol === null || gCol === undefined) {
    var err = new ArangoError();
    err.errorNum = arangodb.errors.ERROR_GRAPH_NO_GRAPH_COLLECTION.code;
    err.errorMessage = arangodb.errors.ERROR_GRAPH_NO_GRAPH_COLLECTION.message;
    throw err;
  }
  return gCol;
};

// //////////////////////////////////////////////////////////////////////////////
// / internal helper to sort a graph's edge definitions
// //////////////////////////////////////////////////////////////////////////////

const sortEdgeDefinition = function (edgeDefinition) {
  edgeDefinition.from = edgeDefinition.from.sort();
  edgeDefinition.to = edgeDefinition.to.sort();
  return edgeDefinition;
};

const checkInfo = (info) => {
  if (!info) {
    let err = new ArangoError();
    err.errorNum = errors.ERROR_GRAPH_NOT_FOUND.code;
    err.errorMessage = errors.ERROR_GRAPH_NOT_FOUND.message;
    throw err;
  }
  if (!info.hasOwnProperty('numberOfShards') ||
      !Number.isInteger(info.numberOfShards) ||
      info.numberOfShards < 0) {
    let err = new ArangoError();
    err.errorNum = errors.ERROR_BAD_PARAMETER.code;
    err.errorMessage = usageMessage;
    throw err;
  }
  if (!info.hasOwnProperty('smartGraphAttribute') ||
      typeof info.smartGraphAttribute !== 'string' ||
      info.smartGraphAttribute === '') {
    let err = new ArangoError();
    err.errorNum = errors.ERROR_BAD_PARAMETER.code;
    err.errorMessage = usageMessage;
    throw err;
  }
  if (info.smartGraphAttribute.startsWith('_')) {
    let err = new ArangoError();
    err.errorNum = errors.ERROR_ILLEGAL_SMART_GRAPH_ATTRIBUTE.code;
    err.errorMessage = errors.ERROR_ILLEGAL_SMART_GRAPH_ATTRIBUTE.message;
    throw err;
  }

  if (info.hasOwnProperty('replicationFactor') &&
      typeof info.replicationFactor !== 'number') {
    let err = new ArangoError();
    err.errorNum = errors.ERROR_BAD_PARAMETER.code;
    err.errorMessage = "The optional parameter 'replicationFactor' has to be an integer number";
    throw err;
  }

  if (info.hasOwnProperty('initial')) {
    // We are from the database
    if (typeof info.initial !== 'string') {
      let err = new ArangoError();
      err.errorNum = errors.ERROR_BAD_PARAMETER.code;
      err.errorMessage = usageMessage;
      throw err;
    }
    if (!info.hasOwnProperty('orphanCollections') || !Array.isArray(info.orphanCollections)) {
      let err = new ArangoError();
      err.errorNum = errors.ERROR_BAD_PARAMETER.code;
      err.errorMessage = usageMessage;
      throw err;
    }
    if (!info.hasOwnProperty('edgeDefinitions') || !Array.isArray(info.edgeDefinitions)) {
      let err = new ArangoError();
      err.errorNum = errors.ERROR_BAD_PARAMETER.code;
      err.errorMessage = usageMessage;
      throw err;
    }
  } else {
    if (info.hasOwnProperty('edgeDefinitions')) {
      if (!Array.isArray(info.edgeDefinitions)) {
        let err = new ArangoError();
        err.errorNum = errors.ERROR_BAD_PARAMETER.code;
        err.errorMessage = usageMessage;
        throw err;
      }
    }
    if (info.hasOwnProperty('orphanCollections')) {
      if (!Array.isArray(info.orphanCollections)) {
        let err = new ArangoError();
        err.errorNum = errors.ERROR_BAD_PARAMETER.code;
        err.errorMessage = usageMessage;
        throw err;
      }
    }
  }

  if (!info.isSmart) {
    let err = new ArangoError();
    err.errorNum = errors.ERROR_GRAPH_INVALID_GRAPH.code;
    err.errorMessage = 'The graph you required is not smart.';
    throw err;
  }

  // If we get here and did not throw the info is good.
};

var createHiddenProperty = function (obj, name, value) {
  Object.defineProperty(obj, name, {
    enumerable: false,
    writable: true
  });
  obj[name] = value;
};

let persist = function (graph) {
  let gdb = getGraphCollection();
  gdb.update(graph.__name, {
    orphanCollections: graph.__orphanCollections,
    edgeDefinitions: graph.__edgeDefinitions
  });
};

class SmartGraph extends gm.__GraphClass {
  constructor (info) {
    super(info);
    createHiddenProperty(this, '__initial', info.initial);
    createHiddenProperty(this, '__numberOfShards', info.numberOfShards);
    createHiddenProperty(this, '__replicationFactor', info.replicationFactor || 1);
    createHiddenProperty(this, '__isSmart', true);
    createHiddenProperty(this, '__smartGraphAttribute', info.smartGraphAttribute);
    createHiddenProperty(this, '__isDisjoint', info.isDisjoint || false);
  }
}

exports._exists = gm._exists;
exports._relation = gm._relation;
exports._edgeDefinitions = gm._edgeDefinitions;
exports._extendEdgeDefinitions = gm._extendEdgeDefinitions;
exports._registerCompatibilityFunctions = gm._registerCompatibilityFunctions;

exports.__SmartGraphClass = SmartGraph;
exports.__GraphClass = gm.__GraphClass;
