/*jshint globalstrict:true, strict:true, esnext: true */

"use strict";

////////////////////////////////////////////////////////////////////////////////
/// DISCLAIMER
///
/// Copyright 2021 ArangoDB GmbH, Cologne, Germany
///
/// Licensed under the Apache License, Version 2.0 (the "License");
/// you may not use this file except in compliance with the License.
/// You may obtain a copy of the License at
///
///     http://www.apache.org/licenses/LICENSE-2.0
///
/// Unless required by applicable law or agreed to in writing, software
/// distributed under the License is distributed on an "AS IS" BASIS,
/// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
/// See the License for the specific language governing permissions and
/// limitations under the License.
///
/// Copyright holder is ArangoDB GmbH, Cologne, Germany
///
/// @author Michael Hackstein
////////////////////////////////////////////////////////////////////////////////

const {db} = require("@arangodb");
const sgm = require("@arangodb/smart-graph");
const {CollectionWrapper} = require("@arangodb/testutils/collection-wrapper-util");

const smartGraphName = "UnitTestHybridSmartGraph";
const satelliteCollectionName = "UnitTestSmartSat";
const verticesCollectionName = "UnitTestSmartVertices";
const edgesCollectionName = "UnitTestSmartEdges";
const edgesSatToSatCollectionName = "UnitTestSatToSatEdges";
const edgesSatToSmartCollectionName = "UnitTestSatToSmartEdges";
const edgesSmartToSatCollectionName = "UnitTestSmartToSatEdges";
const smartGraphAttribute = "smart";
// TODO: Maybe we want to add crazy things like Objects, arrays, numbers, null, undefined here
const validSmartGraphValues = ["B", "a", "C", "LongerName"];

const makeOptions = (isDisjoint) => {
  return {
    satellites: [satelliteCollectionName],
    isDisjoint,
    smartGraphAttribute,
    isSmart: true,
    numberOfShards: 6,
    replicationFactor: 1
  };
};

const createHybridSmartGraph = (isDisjoint) => {
  let options = makeOptions(isDisjoint);
  const edgeDefinitions = [
    sgm._relation(
      edgesCollectionName,
      verticesCollectionName,
      verticesCollectionName
    ),
    sgm._relation(
      edgesSatToSatCollectionName,
      satelliteCollectionName,
      satelliteCollectionName
    ),
    sgm._relation(
      edgesSmartToSatCollectionName,
      verticesCollectionName,
      satelliteCollectionName
    ),
    sgm._relation(
      edgesSatToSmartCollectionName,
      satelliteCollectionName,
      verticesCollectionName
    )
  ];
  sgm._create(smartGraphName, edgeDefinitions, [], options);
};

const dropSmartGraph = () => {
  try {
    sgm._drop(smartGraphName, true);
  } catch (_) {
  }
};

class HybridSmartGraphSatelliteCollectionWrapper extends CollectionWrapper {
  constructor(isDisjoint) {
    super(satelliteCollectionName);
    this._isDisjoint = isDisjoint;
  }

  setUp() {
    createHybridSmartGraph(this._isDisjoint);
  }

  tearDown() {
    dropSmartGraph();
  }
}

function* SmartVertexGenerator(keyGenerator) {
  for (const key of keyGenerator) {
    for (const smartValue of validSmartGraphValues) {
      yield {
        _key: `${smartValue}:${key}`,
        value: "test",
        [smartGraphAttribute]: smartValue
      };
    }
  }
}

class HybridSmartGraphSmartCollectionWrapper extends CollectionWrapper {
  constructor(isDisjoint) {
    super(verticesCollectionName);
    this._isDisjoint = isDisjoint;
  }

  setUp() {
    createHybridSmartGraph(this._isDisjoint);
  }

  tearDown() {
    dropSmartGraph();
  }

  documentGeneratorWithKeys(keyGenerator) {
    return SmartVertexGenerator(keyGenerator);
  }
}

function* SmartEdgeGenerator(keyGenerator, fromIsSmart, toIsSmart, isDisjoint) {
  for (const key of keyGenerator) {
    if (fromIsSmart) {
      for (const fromSmartValue of validSmartGraphValues) {
        const _from = `${verticesCollectionName}/${fromSmartValue}:test`;
        if (toIsSmart) {
          for (const toSmartValue of validSmartGraphValues) {
            if (isDisjoint && fromSmartValue !== toSmartValue) {
              // In the Disjoint case we can only allow _from and _to to be
              // Sharded identically.
              continue;
            }
            const _to = `${verticesCollectionName}/${toSmartValue}:test`;
            yield {
              _key: `${fromSmartValue}:${key}:${toSmartValue}`,
              value: "test",
              _from,
              _to
            };
          }
        } else {
          const _to = `${satelliteCollectionName}/test`;
          yield {
            _key: `${fromSmartValue}:${key}`,
            value: "test",
            _from,
            _to
          };
        }
      }
    } else {
      const _from = `${satelliteCollectionName}/test`;
      if (toIsSmart) {
        for (const toSmartValue of validSmartGraphValues) {
          const _to = `${verticesCollectionName}/${toSmartValue}:test`;
          yield {
            _key: `${key}:${toSmartValue}`,
            value: "test",
            _from,
            _to
          };
        }
      } else {
        const _to = `${satelliteCollectionName}/test`;
        yield {
          _key: `${key}`,
          value: "test",
          _from,
          _to
        };
      }
    }
  }
}

class HybridSmartGraphSmartEdgeCollectionWrapper extends CollectionWrapper {
  constructor(isDisjoint, fromIsSmart, toIsSmart) {
    if (fromIsSmart) {
      if (toIsSmart) {
        super(edgesCollectionName);
      } else {
        super(edgesSmartToSatCollectionName);
      }
    } else {
      if (toIsSmart) {
        super(edgesSatToSmartCollectionName);
      } else {
        super(edgesSatToSatCollectionName);
      }
    }
    this._isDisjoint = isDisjoint;
    this._fromIsSmart = fromIsSmart;
    this._toIsSmart = toIsSmart;
  }

  setUp() {
    createHybridSmartGraph(this._isDisjoint);
  }

  tearDown() {
    dropSmartGraph();
  }

  documentGeneratorWithKeys(keyGenerator) {
    return SmartEdgeGenerator(keyGenerator, this._fromIsSmart, this._toIsSmart, this._isDisjoint);
  }
}

exports.HybridSmartGraphSatelliteCollectionWrapper = HybridSmartGraphSatelliteCollectionWrapper;
exports.HybridSmartGraphSmartCollectionWrapper = HybridSmartGraphSmartCollectionWrapper;
exports.HybridSmartGraphSmartEdgeCollectionWrapper = HybridSmartGraphSmartEdgeCollectionWrapper;